// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

//     **** This file must be preprocessed before compilation ****

// RULE: coordinates on non-static methods are ALWAYS relative to the
// upper-left hand corner of <tt>this</tt>

// FIXME: align
// FIXME use bitfields
// FIXME fixedaspect
// FIXME: reflow before allowing js to read from width/height 
// FIXME: due to font inheritance, we must dirty and mark all null-font descendents of a node if its font changes
// FEATURE: fastpath for rows=1/cols=1
// FEATURE: reflow starting with a certain child
// FEATURE: separate mark_for_reflow and mark_for_resize

import java.io.*;
import java.net.*;
import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;

/**
 *  <p>
 *  Encapsulates the data for a single XWT box as well as all layout
 *  rendering logic.
 *  </p>
 *
 *  <p>
 *  This is the real meat of XWT. Part of its monolithic design is for
 *  performance reasons: deep inheritance heirarchies are slow, and
 *  neither javago nor GCJ can inline across class boundaries.
 *  </p>
 *
 *  <p>The rendering process consists of three phases; each requires
 *     one DFS pass over the tree</p>
 *
 *  <ol><li> <b>repacking</b>: children of a box are packed into columns
 *           and rows according to their colspan/rowspan attributes and
 *           ordering.  Minimum and maximum sizes of columns are computed.
 *
 *      <li> <b>resizing</b>: width/height and x/y positions of children
 *           are assigned.  If a PosChange or SizeChange is triggered,
 *           <tt>Surface.abort</tt> will be set and the resizing process will
 *           Surface.abort.
 *
 *      <li> <b>repainting</b>: children draw their content onto the
 *           buffer.
 *  </ol>
 *
 *  The first two passes together are called the <i>reflow</i> phase.
 *
 *  Reflowing is done in a seperate pass since PosChanges and
 *  SizeChanges trigger an Surface.abort; if rendering were done in the same
 *  pass, rendering work done prior to the Surface.abort would be wasted.
 *
 *  Repacking is seperate from resizing since a box's size depends on
 *  both the box's parent's size (so the traversal must be preorder)
 *  contentwidths of siblings both before and after the box, which in
 *  turn depend on all descendents of the siblings.  FIXME
 *
 *  A note on coordinates: the Box class represents regions
 *  internally as x,y,w,h tuples, even though the DoubleBuffer class
 *  uses x1,y1,x2,y2 tuples.
 */
public final class Box extends JS.Scope {

    public Box() { super(null); }


    // Misc instance data ////////////////////////////////////////////////////////////////

    private static int sizePosChangesSinceLastRender = 0;


    // Misc instance data ////////////////////////////////////////////////////////////////

    boolean needs_reflow = true;         
    //#define MARK_FOR_REFLOW_this for(Box b2 = this; b2 != null && !b2.needs_reflow; b2 = b2.parent) b2.needs_reflow = true;
    //#define MARK_FOR_REFLOW_b for(Box b2 = b; b2 != null && !b2.needs_reflow; b2 = b2.parent) b2.needs_reflow = true;
    //#define MARK_FOR_REFLOW_b_parent for(Box b2 = b.parent; b2 != null && !b2.needs_reflow; b2 = b2.parent) b2.needs_reflow = true;

    private boolean mouseinside = false;
    Box redirect = this;
    Surface surface = null;               // null on all non-root boxen
    Hash traps = null;


    // Geometry ////////////////////////////////////////////////////////////////////////////

    // xwt can be compiled with 16-bit lengths to save memory on small devices
    //#define LENGTH int
    //#define MAX_LENGTH Integer.MAX_VALUE
    //#define MIN_LENGTH Integer.MIN_VALUE

    // always correct (set directly by user)
    LENGTH minwidth = 0;        
    LENGTH minheight = 0;        
    LENGTH maxwidth = MAX_LENGTH;
    LENGTH maxheight = MAX_LENGTH;
    private LENGTH hpad = 0;
    private LENGTH vpad = 0;
    private String text = null;
    private String font = null;
    private LENGTH textwidth = 0;
    private LENGTH textheight = 0;

    // FIXME: use shorts
    private int rows = 1;
    private int cols = 0;
    private int rowspan = 1;
    private int colspan = 1;

    // computed during reflow
    LENGTH x = 0;
    LENGTH y = 0;
    public LENGTH width = 0;
    public LENGTH height = 0;
    private int row = 0;  // FIXME short
    private int col = 0;  // FIXME short
    private LENGTH contentwidth = 0;             // == max(minwidth, textwidth, sum(child.contentwidth) + pad)
    private LENGTH contentheight = 0;


    // Rendering Properties ///////////////////////////////////////////////////////////

    //private SVG.VP path = null;
    //private SVG.Paint fill = null;
    //private SVG.Paint stroke = null;

    private Picture image;                       // will disappear
    private int fillcolor = 0x00000000;          // will become SVG.Paint
    private int strokecolor = 0xFF000000;        // will become SVG.Paint

    private String cursor = null;                // the cursor for this box

    //FIXME make private
    public boolean invisible = false;           // true iff the Box is invisible
    private boolean absolute = false;            // If true, the box will be positioned absolutely
    private boolean vshrink = false;             // If true, the box will shrink to the smallest vertical size possible
    private boolean hshrink = false;             // If true, the box will shrink to the smallest horizontal size possible
    private boolean tile = false; // FIXME: drop this?

    // Instance Methods /////////////////////////////////////////////////////////////////////

    // FIXME: rethink
    /** Adds the intersection of (x,y,w,h) and the node's current actual geometry to the Surface's dirty list */
    public final void dirty() { dirty(0, 0, width, height); }
    public final void dirty(int x, int y, int w, int h) {
        for(Box cur = this; cur != null; cur = cur.parent) {
            w = min(x + w, cur.width) - max(x, 0);
            h = min(y + h, cur.height) - max(y, 0);
            x = max(x, 0);
            y = max(y, 0);
            if (w <= 0 || h <= 0) return;
            if (cur.parent == null && cur.surface != null) cur.surface.dirty(x, y, w, h);
            x += cur.x;
            y += cur.y;
        }
    }

    /**
     *  Given an old and new mouse position, this will update <tt>mouseinside</tt> and check
     *  to see if this node requires any Enter, Leave, or Move notifications.
     *
     *  @param forceleave set to true by the box's parent if the mouse is inside an older
     *                    sibling, which is covering up this box.
     */
    void Move(int oldmousex, int oldmousey, int mousex, int mousey) { Move(oldmousex, oldmousey, mousex, mousey, false); }
    void Move(int oldmousex, int oldmousey, int mousex, int mousey, boolean forceleave) {

        boolean wasinside = mouseinside;
        boolean isinside = !invisible && inside(mousex, mousey) && !forceleave;
        mouseinside = isinside;

        if (!wasinside && !isinside) return;

        if (traps == null) { }
        else if (!wasinside && isinside && traps.get("Enter") != null) put("Enter", Boolean.TRUE);
        else if (wasinside && !isinside && traps.get("Leave") != null) put("Leave", Boolean.TRUE);
        else if (wasinside && isinside && (mousex != oldmousex || mousey != oldmousey) && traps.get("Move") != null) put("Move", Boolean.TRUE);

        if (isinside && cursor != null) getRoot().cursor = cursor;

        // if the mouse has moved into our padding region, it is considered 'outside' all our children
        if (!(mousex >= hpad && mousey >= vpad && mousex < width - hpad && mousey < height + vpad)) forceleave = true;

        for(Box b = getChild(numChildren() - 1); b != null; b = b.prevSibling()) {
            b.Move(oldmousex - b.x, oldmousey - b.y, mousex - b.x, mousey - b.y, forceleave);
            if (b.inside(mousex - b.x, mousey - b.y)) forceleave = true;
        }
    }


    // Reflow ////////////////////////////////////////////////////////////////////////////////////////

    void reflow() {
        repack();
        if (Surface.abort) return;
        resize(x, y, width, height);
    }

    /** Checks if the Box's size has changed, dirties it if necessary, and makes sure childrens' sizes are up to date */
    void repack() {
        if (!needs_reflow) return;
        if (numChildren() == 0) { contentwidth = minwidth; contentheight = minheight; return; }

        // --- Phase 0 ----------------------------------------------------------------------
        // recurse
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            child.repack();
            if (Surface.abort) { MARK_FOR_REFLOW_this; return; }
        }

        // --- Phase 1 ----------------------------------------------------------------------
        // assign children to their row/column positions (assuming constrained columns)
        if ((rows == 0 && cols == 0) || (rows != 0 && cols != 0)) throw new Error("rows == " + rows + "   cols == " + cols);
        //#repeat x/y y/x width/height col/row row/col cols/rows rows/cols colspan/rowspan rowspan/colspan colWidth/rowHeight numRowsInCol/numColsInRow INNER/INNER2 maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight OUTER/OUTER2 INNER/INNER2
        if (rows == 0) {
            int[] numRowsInCol = new int[cols];           // the number of cells occupied in each column
            Box child = getChild(0);
            for(; child != null && (child.absolute || child.invisible); child = child.nextSibling());
            OUTER: for(int row=0; child != null; row++) {
                for(int col=0; child != null && col < cols;) {
                    INNER: while(true) {  // scan across the row, looking for an unoccupied gap at least as wide as the child
                        while(col < cols && numRowsInCol[col] > row) col++;
                        for(int i=col; i < cols && i < col + min(cols, child.colspan); i++)
                            if (numRowsInCol[col] > row) { col = i + 1; continue INNER; }
                        break;
                    }
                    if (col + min(cols, child.colspan) > cols) break;
                    for(int i=col; i < col + min(cols, child.colspan); i++) numRowsInCol[i] += child.rowspan;
                    child.col = col;
                    child.row = row;
                    col += min(cols, child.colspan);
                    child = child.nextSibling();
                    for(; child != null && (child.absolute || child.invisible); child = child.nextSibling());
                }
            }
        }
        //#end

        // --- Phase 2 ----------------------------------------------------------------------
        // compute the min/max sizes of the columns and rows and set our contentwidth
        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight numCols/numRows hpad/vpad
        contentwidth = 2 * hpad;
        int numCols = cols;
        if (numCols == 0)
            for(Box child = getChild(0); child != null; child = child.nextSibling())
                numCols = max(numCols, child.col + child.colspan);
        LENGTH[] colWidth = new LENGTH[numCols];
        for(Box child = getChild(0); child != null; child = child.nextSibling())
            if (!(child.absolute || child.invisible))
                colWidth[child.col] = max(colWidth[child.col], child.contentwidth / child.colspan);
        for(int col=0; col<numCols; col++) contentwidth += colWidth[col];
        contentwidth = max(textwidth + 2 * hpad, contentwidth);
        contentwidth = bound(minwidth, contentwidth, maxwidth);
        //#end
    }


    void resize(LENGTH x, LENGTH y, LENGTH width, LENGTH height) {

        // --- Phase 1 ----------------------------------------------------------------------
        // run PosChange/SizeChange, dirty as needed
        if (x != this.x || y != this.y || width != this.width || height != this.height) {
            (parent == null ? this : parent).dirty(this.x, this.y, this.width, this.height);
            boolean sizechange = false, poschange = false;
            if (traps != null && (this.width != width || this.height != height) && traps.get("SizeChange") != null) sizechange = true;
            if (traps != null && (this.x != x || this.y != y) && traps.get("PosChange") != null) poschange = true;
            this.width = width; this.height = height; this.x = x; this.y = y;
            dirty();
            if (sizechange || poschange)
                if (sizePosChangesSinceLastRender == 500) {
                    if (Log.on) Log.logJS(this, "Warning, more than 500 SizeChange/PosChange traps triggered since last complete render");
                } else {
                    sizePosChangesSinceLastRender++;
                    if (sizechange) put("SizeChange", Boolean.TRUE);
                    if (poschange) put("PosChange", Boolean.TRUE);
                    Surface.abort = true;
                    return;
                }
            needs_reflow = true;
        }

        // --- short circuit ----------------------------------------------------------------
        if (!needs_reflow) return;
        needs_reflow = false;
        if (numChildren() == 0) return;

        // --- Phase 2 ----------------------------------------------------------------------
        // compute the min/max sizes of the columns and rows and set initial width/height to minimums

        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight marginWidth/marginHeight numCols/numRows
        int numCols = cols;
        if (numCols == 0)
            for(Box child = getChild(0); child != null; child = child.nextSibling())
                numCols = max(numCols, child.col + child.colspan);
        LENGTH[] colWidth = new LENGTH[numCols];
        LENGTH[] colMaxWidth = new LENGTH[numCols];
        int marginWidth = width;
        for(int i=0; i<colMaxWidth.length; i++) colMaxWidth[i] = -1;
        //#end

        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (child.absolute || child.invisible) continue;
            //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight hshrink/vshrink numCols/numRows
            colWidth[child.col] = max(colWidth[child.col], child.contentwidth / child.colspan);
            for(int i=child.col; i<child.col+child.colspan && i<numCols; i++)
                colMaxWidth[i] = max(colMaxWidth[i], (child.hshrink ? child.contentwidth : child.maxwidth) / child.colspan);
            //#end
        }

        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight marginWidth/marginHeight
        for(int i=0; i<colMaxWidth.length; i++) if (colMaxWidth[i] == -1) colMaxWidth[i] = MAX_LENGTH;

        for(int i=0; i<colMaxWidth.length; i++) {
            if (colMaxWidth[i] == MAX_LENGTH) { marginWidth = 0; break; }
            marginWidth -= colMaxWidth[i];
            if (marginWidth < 0) { marginWidth = 0; break; }
        }
        //#end
      

        // --- Phase 3 ----------------------------------------------------------------------
        // hand out the slack
        int slack;
        //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight numCols/numRows
        slack = width;
        for(int i=0; i<numCols; i++) slack -= colWidth[i];
        if (numChildren() > 0)
            while(slack > 0) {  
                // FEATURE: inefficient
                int startslack = slack;
                int increment = max(1, slack / numCols);
                for(int col=0; col < numCols && slack > 0; col++) {
                    slack += colWidth[col];
                    colWidth[col] = min(colMaxWidth[col], colWidth[col] + increment);
                    slack -= colWidth[col];
                }
                if (slack == startslack) break;
            }   
        //#end


        // --- Phase 4 ----------------------------------------------------------------------
        // assign children's new sizes and positions and recurse
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (child.invisible) continue;
            int child_x = 0, child_y = 0, child_width = 0, child_height = 0;
            if (child.absolute) {
                child_x = child.x;
                child_y = child.y;
                child_width = child.hshrink ? child.contentwidth : min(child.maxwidth, width - child.x - hpad);
                child_height = child.vshrink ? child.contentheight : min(child.maxheight, height - child.y - vpad);
            } else {
                int diff;
                //#repeat x/y y/x width/height col/row cols/rows colspan/rowspan colWidth/rowHeight maxwidth/maxheight minwidth/minheight contentwidth/contentheight colMaxWidth/rowMaxHeight hshrink/vshrink marginWidth/marginHeight hpad/vpad child_x/child_y child_width/child_height
                child_width = 0; for(int i=child.col; i<child.col+child.colspan && i<colWidth.length; i++) child_width += colWidth[i];
                diff = bound(child.contentwidth, child_width, child.hshrink ? child.contentwidth : child.maxwidth) - child_width;
                child_x = max(hpad, marginWidth / 2); for(int i=0; i<child.col; i++) child_x += colWidth[i];
                if (diff < 0) child_x += -1 * (diff / 2);
                child_width += diff;
                //#end
            }
            child.resize(child_x, child_y, child_width, child_height);
        }
    }




    // Rendering Pipeline /////////////////////////////////////////////////////////////////////

    /** Renders self and children within the specified region. All rendering operations are clipped to xIn,yIn,wIn,hIn */
    void render(int parentx, int parenty, int clipx, int clipy, int clipw, int cliph, DoubleBuffer buf) {
        if (Surface.abort || invisible) return;
        int globalx = parentx + (parent == null ? 0 : x);
        int globaly = parenty + (parent == null ? 0 : y);

        // intersect the x,y,w,h rendering window with ourselves; quit if it's empty
        clipw = min(max(clipx, parent == null ? 0 : globalx) + clipw, (parent == null ? 0 : globalx) + width) - globalx;
        cliph = min(max(clipy, parent == null ? 0 : globaly) + cliph, (parent == null ? 0 : globaly) + height) - globaly;
        clipx = max(clipx, parent == null ? 0 : globalx);
        clipy = max(clipy, parent == null ? 0 : globaly);
        if (clipw <= 0 || cliph <= 0) return;

        if ((fillcolor & 0xFF000000) != 0x00000000 || parent == null)
            buf.fillRect(clipx, clipy, clipx + clipw, clipy + cliph, (fillcolor & 0xFF000000) != 0 ? fillcolor : 0xFF777777);

        if (image != null)
            if (tile) renderTiledImage(globalx, globaly, clipx, clipy, clipw, cliph, buf);
            else renderStretchedImage(globalx, globaly, clipx, clipy, clipw, cliph, buf);

        //if (text != null && !text.equals("")) renderText(x, y, w, h, buf);

        // now subtract the pad region from the clip region before proceeding
        clipw = min(max(clipx, globalx + hpad) + clipw, globalx + width - hpad) - clipx;
        cliph = min(max(clipy, globaly + vpad) + cliph, globaly + height - vpad) - clipy;
        clipx = max(clipx, globalx + hpad);
        clipy = max(clipy, globaly + vpad);

        for(Box b = getChild(0); b != null; b = b.nextSibling()) b.render(globalx, globaly, clipx, clipy, clipw, cliph, buf);   
    }

    void renderStretchedImage(int globalx, int globaly, int x, int y, int w, int h, DoubleBuffer buf) {
        buf.setClip(x, y, w + x, h + y);

        /*
        if (fixedaspect) {
            int hstretch = width / image.getWidth();
            if (hstretch == 0) hstretch = -1 * image.getWidth() / width;
            int vstretch = height / image.getHeight();
            if (vstretch == 0) vstretch = -1 * image.getHeight() / height;
            if (hstretch < vstretch) height = image.getHeight() * width / image.getWidth();
            else width = image.getWidth() * height / image.getHeight();
        }
        */

        buf.drawPicture(image, globalx, globaly, globalx + width, globaly + height, 0, 0, image.getWidth(), image.getHeight());
        buf.setClip(0, 0, buf.getWidth(), buf.getHeight());
    }

    void renderTiledImage(int globalx, int globaly, int x, int y, int w, int h, DoubleBuffer buf) {
        int iw = image.getWidth();
        int ih = image.getHeight();
        // FIXME broken
        for(int i=(x - x)/iw; i <= (x + w - x)/iw; i++) {
            for(int j=(y - y)/ih; j<= (y + h - y)/ih; j++) {
                
                int dx1 = max(i * iw + x, x);
                int dy1 = max(j * ih + y, y);
                int dx2 = min((i+1) * iw + x, x + w);
                int dy2 = min((j+1) * ih + y, y + h);
                
                int sx1 = dx1 - (i*iw) - x;
                int sy1 = dy1 - (j*ih) - y;
                int sx2 = dx2 - (i*iw) - x;
                int sy2 = dy2 - (j*ih) - y;

                if (dx2 - dx1 > 0 && dy2 - dy1 > 0 && sx2 - sx1 > 0 && sy2 - sy1 > 0)
                    buf.drawPicture(image, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2);
            }
        }
    }

    void renderText(int x, int y, int w, int h, DoubleBuffer buf) {
        /*
        if ((textcolor & 0xFF000000) == 0x00000000) return;
        buf.setClip(x, y, w + x, h + y);

        buf.drawString(font(), text, x + hpad, y + vpad + Platform.getMaxAscent(font()) - 1, textcolor);
        buf.setClip(0, 0, buf.getWidth(), buf.getHeight());
        int i=0; while(i<font().length() && !Character.isDigit(font().charAt(i))) i++;

        if (font().lastIndexOf('d') > i) {
            for(int j = x + hpad; j < x + hpad + textdim(0); j += 2)
                buf.fillRect(j, y + vpad + (xwf == null ? Platform.getMaxAscent(font()) : xwf.getMaxAscent()) + 2,
                             j + 1, y + vpad + (xwf == null ? Platform.getMaxAscent(font()) : xwf.getMaxAscent()) + 2 + 1,
                             textcolor);

        } else if (font().lastIndexOf('u') > i) {
            buf.fillRect(x + hpad,
                        y + vpad + (xwf == null ? Platform.getMaxAscent(font()) : xwf.getMaxAscent()) + 2,
                        x + hpad + textdim(0),
                        y + vpad + (xwf == null ? Platform.getMaxAscent(font()) : xwf.getMaxAscent()) + 2 + 1,
                        textcolor);
        }
        */
    }


    // Methods to implement org.xwt.js.JS //////////////////////////////////////

    public Object callMethod(Object method, JS.Array args, boolean checkOnly) throws JS.Exn {
        if ("indexof".equals(method)) {
            if (checkOnly) return Boolean.TRUE;
            if (args.length() != 1 || args.elementAt(0) == null || !(args.elementAt(0) instanceof Box)) return new Integer(-1);
            Box b = (Box)args.elementAt(0);
            if (b.parent != Box.this) {
                if (redirect == null || redirect == Box.this) return new Integer(-1);
                return redirect.callMethod(method, args, checkOnly);
            }
            return new Integer(b.getIndexInParent());

        } else if ("apply".equals(method)) {
            if (checkOnly) return Boolean.TRUE;
            if (args.elementAt(0) instanceof String) {
                String templatename = (String)args.elementAt(0);
                Template t = Template.getTemplate(templatename, null);
                if (t == null) {
                    if (Log.on) Log.logJS(this, "template " + templatename + " not found");
                } else {
                    if (ThreadMessage.suspendThread()) try {
                        JS.Callable callback = args.length() < 2 ? null : (Callable)args.elementAt(1);
                        t.apply(this, null, null, callback, 0, t.numUnits());
                    } finally {
                        ThreadMessage.resumeThread();
                    }
                }
            } else if (args.elementAt(0) instanceof JS && !(args.elementAt(0) instanceof Box)) {
                JS s = (JS)args.elementAt(0);
                Object[] keys = s.keys();
                for(int j=0; j<keys.length; j++) put(keys[j].toString(), s.get(keys[j]));
            }
            return this;
        }
        return null;
    }

    /** Returns the i_th child */
    public Object get(int i) {
        if (redirect == null) return null;
        if (redirect != this) return redirect.get(i);
        return i >= numChildren() || i < 0 ? null : getChild(i);
    }

    /**
     *  Inserts value as child i; calls remove() if necessary.
     *  This method handles "reinserting" one of your children properly.
     *  INVARIANT: after completion, getChild(min(i, numChildren())) == newnode
     *  WARNING: O(n) runtime, unless i == numChildren()
     */
    public void put(int i, Object value) {
        if (i < 0) return;

        if (value != null && !(value instanceof Box)) {
            if (Log.on) Log.logJS(this, "attempt to set a numerical property on a box to anything other than a box");
        } else if (redirect == null) {
            if (Log.on) Log.logJS(this, "attempt to add/remove children to/from a node with a null redirect");
        } else if (redirect != this) {
            Box b = value == null ? (Box)redirect.get(i) : (Box)value;
            redirect.put(i, value);
            put("0", b);
        } else if (value == null) {
            if (i >= 0 && i < numChildren()) {
                Box b = getChild(i);
                b.remove();
                put("0", b);
            }
        } else if (value instanceof RootProxy) {
            if (Log.on) Log.logJS(this, "attempt to reparent a box via its proxy object");
        } else {
            Box newnode = (Box)value;

            // check if box being moved is currently target of a redirect
            for(Box cur = newnode.parent; cur != null; cur = cur.parent)
                if (cur.redirect == newnode) {
                    if (Log.on) Log.logJS(this, "attempt to move a box that is the target of a redirect");
                    return;
                }

            // check for recursive ancestor violation
            for(Box cur = this; cur != null; cur = cur.parent)
                if (cur == newnode) {
                    if (Log.on) Log.logJS(this, "attempt to make a node a parent of its own ancestor");
                    if (Log.on) Log.log(this, "box == " + this + "  ancestor == " + newnode);
                    return;
                }

            if (numKids > 15 && children == null) convert_to_array();
            newnode.remove();
            newnode.parent = this;
            
            if (children == null) {
                if (firstKid == null) {
                    firstKid = newnode;
                    newnode.prevSibling = newnode;
                    newnode.nextSibling = newnode;
                } else if (i >= numKids) {
                    newnode.prevSibling = firstKid.prevSibling;
                    newnode.nextSibling = firstKid;
                    firstKid.prevSibling.nextSibling = newnode;
                    firstKid.prevSibling = newnode;
                } else {
                    Box cur = firstKid;
                    for(int j=0; j<i; j++) cur = cur.nextSibling;
                    newnode.prevSibling = cur.prevSibling;
                    newnode.nextSibling = cur;
                    cur.prevSibling.nextSibling = newnode;
                    cur.prevSibling = newnode;
                    if (i == 0) firstKid = newnode;
                }
                numKids++;
                
            } else {
                if (i >= children.size()) {
                    newnode.indexInParent = children.size();
                    children.addElement(newnode);
                } else {
                    children.insertElementAt(newnode, i);
                    for(int j=i; j<children.size(); j++)
                        getChild(j).indexInParent = j;
                }
            }
            
            // need both of these in case child was already uncalc'ed
            Box b = newnode; 
            MARK_FOR_REFLOW_b;
            MARK_FOR_REFLOW_this;
            
            newnode.dirty();

            // note that JavaScript box[0] will invoke put(int i), not put(String s)
            put("0", newnode);
        }
    }
    
    public Object get(Object name) { return get(name, false); }
    public Object get(Object name_, boolean ignoretraps) {
        if (name_ instanceof Number) return get(((Number)name_).intValue());

        if (!(name_ instanceof String)) return null;
        String name = (String)name_;
        if (name.equals("")) return null;

        // See if we're reading back the function value of a trap
        if (name.charAt(0) == '_') {
            if (name.charAt(1) == '_') name = name.substring(2);
            else name = name.substring(1);
            Trap t = Trap.getTrap(this, name);
            return t == null ? null : t.f;
        }
        
        // See if we're triggering a trap
        Trap t = traps == null || ignoretraps ? null : (Trap)traps.get(name);
        if (t != null && t.isreadtrap) return t.perform(Trap.emptyargs);

        // Check for a special handler
        SpecialBoxProperty gph = (SpecialBoxProperty)SpecialBoxProperty.specialBoxProperties.get(name);
        if (gph != null) return gph.get(this);

        Object ret = super.get(name);
        if (name.startsWith("$") && ret == null)
            if (Log.on) Log.logJS(this, "WARNING: attempt to access " + name + ", but no child with id=\"" + name.substring(1) + "\" found");
        return ret;
    }

    public Object[] keys() {
        Object[] ret = new Object[numChildren()];
        for(int i=0; i<ret.length; i++) ret[i] = new Integer(i);
        return ret;
    }

    /**
     *  Scriptable.put()
     *  @param ignoretraps if set, no traps will be triggered (set when 'cascade' reaches the bottom of the trap stack)
     *  @param rp if this put is being performed via a root proxy, rp is the root proxy.
     */
    public void put(Object name, Object value) { put(name, value, false, null); }
    public void put(Object name, Object value, boolean ignoretraps) { put(name, value, ignoretraps, null); }
    public void put(Object name_, Object value, boolean ignoretraps, RootProxy rp) {
        if (name_ instanceof Number) { put(((Number)name_).intValue(), value); return; }
        if (!(name_ instanceof String)) { super.put(name_,value); return; }
        String name = name_.toString();
        if (!ignoretraps && traps != null) {
            Trap t = (Trap)traps.get(name);
            if (t != null) {
                JS.Array arg = new JS.Array();
                arg.addElement(value);
                t.perform(arg);
                arg.setElementAt(null, 0);
                return;
            }
        }

        // don't want to really cascade down to the box on this one
        if (name.equals("0")) return;

        SpecialBoxProperty gph = (SpecialBoxProperty)SpecialBoxProperty.specialBoxProperties.get(name);
        if (gph != null) { gph.put(name, this, value); return; }

        if (name.charAt(0) == '_') {
            if (value != null && !(value instanceof JS.Callable)) {
                if (Log.on) Log.logJS(this, "attempt to put a non function value (" + value + ") to " + name);
            } else if (value != null && !(value instanceof JS.CompiledFunction)) {
                if (Log.on) Log.logJS(this, "attempt to put a non-compiled function value (" + value + ") to " + name);
            } else if (name.charAt(1) == '_') {
                name = name.substring(2).intern();
                Trap t = Trap.getTrap(this, name);
                if (t != null) t.delete();
                if (value != null) Trap.addTrap(this, name, ((JS.CompiledFunction)value), true, rp);
            } else {
                name = name.substring(1).intern();
                Trap t = Trap.getTrap(this, name);
                if (t != null) t.delete();
                if (value != null) Trap.addTrap(this, name, ((JS.CompiledFunction)value), false, rp);
            }
            return;
        }

        super.put(name, value);
    }


    // Tree Manipulation /////////////////////////////////////////////////////////////////////

    /** The parent of this node */
    private Box parent = null;
    
    // Variables used in Vector mode */
    /** INVARIANT: if (parent != null) parent.children.elementAt(indexInParent) == this */
    private int indexInParent;
    private Vec children = null;

    // Variables used in linked-list mode
    private int numKids = 0;
    private Box nextSibling = null;
    private Box prevSibling = null;
    private Box firstKid = null;
    
    // when we get more than 15 children, we switch to array-mode
    private void convert_to_array() {
        children = new Vec(numKids);
        Box cur = firstKid;
        do {
            children.addElement(cur);
            cur.indexInParent = children.size() - 1;
            cur = cur.nextSibling;
        } while (cur != firstKid);
    }
    
    /** remove this node from its parent; INVARIANT: whenever the parent of a node is changed, remove() gets called. */
    public void remove() {
        if (parent == null) {
            if (surface != null) surface.dispose(true);
            return;
        }
        Box oldparent = parent;
        if (oldparent == null) return;
        MARK_FOR_REFLOW_this;
        dirty();
        mouseinside = false;

        if (parent.children != null) {
            parent.children.removeElementAt(indexInParent);
            for(int j=indexInParent; j<parent.children.size(); j++)
                (parent.getChild(j)).indexInParent = j;

        } else {
            if (parent.firstKid == this) {
                if (nextSibling == this) parent.firstKid = null;
                else parent.firstKid = nextSibling;
            }
            parent.numKids--;
            prevSibling.nextSibling = nextSibling;
            nextSibling.prevSibling = prevSibling;
            prevSibling = null;
            nextSibling = null;
        }
        parent = null;

        if (oldparent != null) { Box b = oldparent; MARK_FOR_REFLOW_b; }

        // note that JavaScript box[0] will invoke put(int i), not put(String s)
        if (oldparent != null) oldparent.put("0", this);
    }

    /** returns our next sibling (parent[ourindex + 1]) */
    public final Box nextSibling() {
        if (parent == null) return null;
        if (parent.children == null) {
            if (nextSibling == parent.firstKid) return null;
            return nextSibling;
        } else {
            if (indexInParent >= parent.children.size() - 1) return null;
            return (Box)parent.children.elementAt(indexInParent + 1);
        }
    }
    
    /** returns our next sibling (parent[ourindex + 1]) */
    public final Box prevSibling() {
        if (parent == null) return null;
        if (parent.children == null) {
            if (this == parent.firstKid) return null;
            return prevSibling;
        } else {
            if (indexInParent == 0) return null;
            return (Box)parent.children.elementAt(indexInParent - 1);
        }
    }
    
    /** Returns the parent of this node */
    public Box getParent() { return parent; }
    
    /** Returns ith child */
    public Box getChild(int i) {
        if (children == null) {
            if (firstKid == null) return null;
            if (i >= numKids) return null;
            if (i == numKids - 1) return firstKid.prevSibling;
            Box cur = firstKid;
            for(int j=0; j<i; j++) cur = cur.nextSibling;
            return cur;
        } else {
            if (i >= children.size() || i < 0) return null;
            return (Box)children.elementAt(i);
        }
    }
    
    /** Returns the number of children */
    public int numChildren() {
        if (children == null) {
            if (firstKid == null) return 0;
            int i=1;
            for(Box cur = firstKid.nextSibling; cur != firstKid; i++) cur = cur.nextSibling;
            return i;
        } else {
            return children.size();
        }
    }
    
    /** Returns our index in our parent */
    public int getIndexInParent() {
        if (parent == null) return 0;
        if (parent.children == null) {
            int i = 0;
            for(Box cur = this; cur != parent.firstKid; i++) cur = cur.prevSibling;
            return i;
        } else {
            return indexInParent;
        }
    }

    /** returns the root of the surface that this box belongs to */
    public final Box getRoot() {
        if (parent == null && surface != null) return this;
        if (parent == null) return null;
        return parent.getRoot();
    }


    // Root Proxy ///////////////////////////////////////////////////////////////////////////////

    // FEATURE: use xwt.graft() here
    RootProxy myproxy = null;
    public JS getRootProxy() {
        if (myproxy == null) myproxy = new RootProxy(this);
        return myproxy;
    }

    private static class RootProxy extends JS {
        Box box;
        RootProxy(Box b) { this.box = b; }
        public Object get(Object name) { return box.get(name); }
        public void put(Object name, Object value) { box.put(name, value, false, this); }
        public Object[] keys() { return box.keys(); }
        public Object callMethod(Object method, JS.Array args, boolean justChecking) {
	    return ((Box)box).callMethod(method,args,justChecking);
	}
    }


    // Trivial Helper Methods (should be inlined) /////////////////////////////////////////

    static final int min(int a, int b) { if (a<b) return a; else return b; }
    static final double min(double a, double b) { if (a<b) return a; else return b; }
    static final int max(int a, int b) { if (a>b) return a; else return b; }
    static final int min(int a, int b, int c) { if (a<=b && a<=c) return a; else if (b<=c && b<=a) return b; else return c; }
    static final int max(int a, int b, int c) { if (a>=b && a>=c) return a; else if (b>=c && b>=a) return b; else return c; }
    static final int bound(int a, int b, int c) { if (c < b) return c; if (a > b) return a; return b; }
    final boolean inside(int x, int y) { return (!invisible && x >= 0 && y >= 0 && x < width && y < height); }
    
    /** figures out what box in this subtree of the Box owns the pixel at x,y relitave to the Surface */
    public static Box whoIs(Box cur, int x, int y) {

        if (cur.parent != null) throw new Error("whoIs may only be invoked on the root box of a surface");
        int globalx = 0;
        int globaly = 0;

        // WARNING: this method is called from the event-queueing
        // thread -- it may run concurrently with ANY part of XWT, and
        // is UNSYNCHRONIZED for performance reasons.  BE CAREFUL
        // HERE.

        if (cur.invisible) return null;
        if (!cur.inside(x - globalx, y - globaly)) return cur.parent == null ? cur : null;
        OUTER: while(true) {
            for(int i=cur.numChildren() - 1; i>=0; i--) {
                Box child = cur.getChild(i);
                if (child == null) continue;        // since this method is unsynchronized, we have to double-check
                globalx += child.x;
                globaly += child.y;
                if (!child.invisible && child.inside(x - globalx, y - globaly)) { cur = child; continue OUTER; }
                globalx -= child.x;
                globaly -= child.y;
            }
            break;
        }
        return cur;
    }

    /** 
     *  A helper class for properties of Box which require special
     *  handling.
     *
     *  To avoid excessive use of String.equals(), the Box.get() and
     *  Box.put() methods employ a Hash keyed on property names that
     *  require special handling. The value stored in the Hash is an
     *  instance of an anonymous subclass of SpecialBoxProperty, which knows
     *  how to handle get()s and put()s for that property name. There
     *  should be one anonymous subclass of SpecialBoxProperty for each
     *  specially-handled property on Box.
     */
    static class SpecialBoxProperty {

        SpecialBoxProperty() { }

        /** stores instances of SpecialBoxProperty; keyed on property name */
        static Hash specialBoxProperties = new Hash(200, 3);

        /** this method defines the behavior when the property is get()ed from b */
        Object get(Box b) { return null; }

        /** this method defines the behavior when the property is put() to b */
        void put(Box b, Object value) { }

        /** this method defines the behavior when the property is put() to b, allows a single SpecialBoxProperty to serve multiple properties */
        void put(String name, Box b, Object value) { put(b, value); }

        static {
            //#repeat fillcolor/strokecolor
            specialBoxProperties.put("fillcolor", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if ((b.fillcolor & 0xFF000000) == 0) return null;
                        String red = Integer.toHexString((b.fillcolor & 0x00FF0000) >> 16);
                        String green = Integer.toHexString((b.fillcolor & 0x0000FF00) >> 8);
                        String blue = Integer.toHexString(b.fillcolor & 0x000000FF);
                        if (red.length() < 2) red = "0" + red;
                        if (blue.length() < 2) blue = "0" + blue;
                        if (green.length() < 2) green = "0" + green;
                        return "#" + red + green + blue;
                    }
                    public void put(Box b, Object value) {
                        int newcolor = b.fillcolor;
                        String s = value == null ? null : value.toString();
                        if (value == null) newcolor = 0x00000000;
                        else if (s.length() > 0 && s.charAt(0) == '#')
                            try {
                                newcolor = 0xFF000000 |
                                    (Integer.parseInt(s.substring(1, 3), 16) << 16) |
                                    (Integer.parseInt(s.substring(3, 5), 16) << 8) |
                                    Integer.parseInt(s.substring(5, 7), 16);
                            } catch (NumberFormatException e) {
                                Log.log(this, "invalid color " + s);
                                return;
                            }
                        else if (SVG.colors.get(s) != null)
                            newcolor = 0xFF000000 | ((Integer)SVG.colors.get(s)).intValue();
                        // FIXME put named colors back in
                        if (newcolor == b.fillcolor) return;
                        b.fillcolor = newcolor;
                        b.dirty();
                    }
                });
            //#end
        
            specialBoxProperties.put("color", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.get("fillcolor"); }
                    public void put(Box b, Object value) { b.put("fillcolor", value); }
                });
            specialBoxProperties.put("textcolor", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.get("strokecolor"); }
                    public void put(Box b, Object value) { b.put("strokecolor", value); }
                });

            specialBoxProperties.put("text", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.text; }
                    public void put(Box b, Object value) {
                        String t = value == null ? "null" : value.toString();
                        if (t.equals(b.text)) return;
                        // FIXME text is broken
                    } });
            specialBoxProperties.put("font", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.font; }
                    public void put(Box b, Object value) {
                        b.font = value == null ? null : value.toString();
                        //b.fontChanged();
                        b.dirty();
                    } });
        
            specialBoxProperties.put("thisbox", new SpecialBoxProperty() {
                    public Object get(Box b) { return b; }
                    public void put(Box b, Object value) {
                        if (value == null) b.remove();
                        else if (value.equals("window") || value.equals("frame")) Platform.createSurface(b, value.equals("frame"), true);
                        else if (Log.on) Log.log(this, "put invalid value to 'thisbox' property: " + value);
                    }
                });

            specialBoxProperties.put("orient", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.redirect == null) return "horizontal";
                        else if (b.redirect != b) return get(b.redirect);
                        else if (b.cols == 1) return "vertical";
                        else if (b.rows == 1) return "horizontal";
                        else return "grid";
                    }
                    public void put(Box b, Object value) {
                        if (value == null) return;
                        if (b.redirect == null) return;
                        if (b.redirect != b) { put(b.redirect, value); return; }
                        if (value.equals("vertical")) {
                            if (b.rows == 0) return;
                            b.rows = 0; b.cols = 1;
                        } else if (value.equals("horizontal")) {
                            if (b.cols == 0) return;
                            b.cols = 0; b.rows = 1;
                        } else if (Log.on)
                            Log.log(this, "invalid value put to orient property: " + value);
                        MARK_FOR_REFLOW_b;
                    } });

            specialBoxProperties.put("static", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        String cfsn =
                            JS.Thread.fromJavaThread(java.lang.Thread.currentThread()).getCurrentCompiledFunction().getSourceName();
                        for(int i=0; i<cfsn.length() - 1; i++)
                            if (cfsn.charAt(i) == '.' && (cfsn.charAt(i+1) == '_' || Character.isDigit(cfsn.charAt(i+1)))) {
                                cfsn = cfsn.substring(0, i);
                                break;
                            }
                        return Static.getStatic(cfsn);
                    }
                });

            specialBoxProperties.put("shrink", new SpecialBoxProperty() {
                    public Object get(Box b) { return (b.vshrink && b.hshrink) ? Boolean.TRUE : Boolean.FALSE; }
                    public void put(Box b, Object value) { b.put("hshrink", value); b.put("vshrink", value); }
                });
        
            //#repeat hshrink/vshrink
            specialBoxProperties.put("hshrink", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Boolean(b.hshrink); }
                    public void put(Box b, Object value) {
                        boolean newshrink = stob(value);
                        if (b.hshrink == newshrink) return;
                        b.hshrink = newshrink;
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end
        
            //#repeat x/y
            specialBoxProperties.put("x", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.surface == null) return new Integer(0);
                        if (b.invisible) return new Integer(0);
                        return new Integer(b.x);
                    }
                    public void put(Box b, Object value) {
                        if (!b.absolute) return;
                        int x = stoi(value);
                        if (x == b.x) return;
                        b.dirty();
                        b.x = x;
                        if (b.parent == null && b.surface != null) {
                            // FIXME this gets hosed by the #repeat
                            //b.surface.setLocation(b.x, b.y);
                            b.surface.centerSurfaceOnRender = false;
                        }
                        MARK_FOR_REFLOW_b;
                        b.dirty();
                    }
                });
            //#end
        
            //#repeat width/height minwidth/minheight maxwidth/maxheight
            specialBoxProperties.put("width", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.width); }
                    public void put(Box b, Object value) {
                        b.width = stoi(value);
                        if (b.parent == null && b.surface != null) {
                            b.surface.setSize();
                            MARK_FOR_REFLOW_b;
                        } else {
                            b.minwidth = b.maxwidth = b.width;
                            MARK_FOR_REFLOW_b;
                        }
                    } });
            //#end

            //#repeat cols/rows rows/cols
            specialBoxProperties.put("cols", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Double(b.cols); }
                    public void put(Box b, Object value) {
                        b.cols = stoi(value);
                        if (b.cols == 0 && b.rows == 0) b.rows = 1;
                        if (b.cols != 0 && b.rows != 0) b.rows = 0;
                        MARK_FOR_REFLOW_b;
                    } });
            //#end
        
            //#repeat colspan/rowspan
            specialBoxProperties.put("colspan", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Double(b.colspan); }
                    public void put(Box b, Object value) { b.colspan = stoi(value); MARK_FOR_REFLOW_b; }
                });
            //#end
        
            specialBoxProperties.put("tile", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.tile ? Boolean.TRUE : Boolean.FALSE; }
                    public void put(Box b, Object value) {
                        boolean newtile = stob(value);
                        if (newtile == b.tile) return;
                        b.tile = newtile;
                        b.dirty();
                    } });
        
            specialBoxProperties.put("invisible", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        for (Box cur = b; cur != null; cur = cur.parent) { if (cur.invisible) return Boolean.TRUE; }
                        return Boolean.FALSE;
                    }
                    public void put(Box b, Object value) {
                        boolean newinvisible = stob(value);
                        if (newinvisible == b.invisible) return;
                        b.invisible = newinvisible;
                        if (b.parent == null) {
                            if (b.surface != null) b.surface.setInvisible(newinvisible);
                        } else {
                            b.dirty();
                            MARK_FOR_REFLOW_b_parent;
                            b.parent.dirty(b.x, b.y, b.width, b.height);
                        }
                    }});
        
            specialBoxProperties.put("absolute", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.absolute ? Boolean.TRUE : Boolean.FALSE; }
                    public void put(Box b, Object value) {
                        boolean newabsolute = stob(value);
                        if (newabsolute == b.absolute) return;
                        b.absolute = newabsolute;
                        if (b.absolute) { b.x = 0; b.y = 0; }
                        if (b.parent != null) MARK_FOR_REFLOW_b_parent;
                    } });
        
            specialBoxProperties.put("image", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.image == null ? null : ImageDecoder.imageToNameMap.get(b.image); }
                    public void put(Box b, Object value) {
                        if ((value == null && b.image == null) ||
                            (value != null && b.image != null && value.equals(ImageDecoder.imageToNameMap.get(b.image)))) return;
                        String s = value == null ? null : value.toString();
                        if (s == null || s.equals("")) b.image = null;
                        else {
                            if ((b.image = ImageDecoder.getPicture(s)) == null) {
                                if (Log.on) Log.logJS(Box.class, "unable to load image " + s);
                            } else {
                                b.minwidth = b.maxwidth = b.image.getWidth();
                                b.minheight = b.maxheight = b.image.getHeight();
                                MARK_FOR_REFLOW_b;
                            }
                        }
                        b.dirty();
                    }
                });

            //#repeat globalx/globaly x/y
            specialBoxProperties.put("globalx", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.parent == null || b.surface == null ? 0 : b.x); }
                    public void put(Box b, Object value) {
                        if (b.surface == null || b.parent == null) return;
                        b.put("x", new Integer(stoi(value) - stoi(get(b.parent))));
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end
        
            specialBoxProperties.put("cursor", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.cursor; } 
                    public void put(Box b, Object value) {
                        b.cursor = (String)value;
                        if (b.surface == null) return;

                        // see if we need to update the surface cursor
                        Surface surface = b.getRoot().surface;
                        String tempcursor = surface.cursor;
                        b.Move(surface.mousex, surface.mousey, surface.mousex, surface.mousey);
                        if (surface.cursor != tempcursor) surface.syncCursor();
                    } 
                });
        
            //#repeat mousex/mousey x/y
            specialBoxProperties.put("mousex", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        Surface surface = b.getRoot().surface;
                        if (surface == null) return new Integer(0);
                        int mousex = surface.mousex;
                        for(Box cur = b; cur != null && cur.parent != null; cur = cur.parent) mousex -= cur.x;
                        return new Integer(mousex);
                    }
                });
            //#end
        
            specialBoxProperties.put("xwt", new SpecialBoxProperty() {
                    public Object get(Box b) { return XWT.singleton; }
                });
        
            specialBoxProperties.put("mouseinside", new SpecialBoxProperty() {
                    public Object get(Box b) { return b.mouseinside ? Boolean.TRUE : Boolean.FALSE; }
                });
        
            specialBoxProperties.put("numchildren", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.redirect == null) return new Integer(0);
                        if (b.redirect != b) return get(b.redirect);
                        return new Integer(b.numChildren());
                    } });
        
            SpecialBoxProperty mouseEventHandler = new SpecialBoxProperty() {
                    public void put(String name, Box b, Object value) {
                        Surface surface = b.getRoot().surface;
                        if (surface == null) return;
                        int mousex = surface.mousex;
                        int mousey = surface.mousey;
                        for(Box c = b.parent; c != null && c.parent != null; c = c.parent) {
                            mousex -= c.x;
                            mousey -= c.y;
                        }
                        for(Box c = b.prevSibling(); c != null; c = c.prevSibling()) {
                            if (c.inside(mousex - c.x, mousey - c.y)) {
                                c.put(name, value);
                                return;
                            }
                        }
                        if (b.parent != null) b.parent.put(name, value);
                    }};

            specialBoxProperties.put("Press1", mouseEventHandler);
            specialBoxProperties.put("Press2", mouseEventHandler);
            specialBoxProperties.put("Press3", mouseEventHandler);
            specialBoxProperties.put("Release1", mouseEventHandler);
            specialBoxProperties.put("Release2", mouseEventHandler);
            specialBoxProperties.put("Release3", mouseEventHandler);
            specialBoxProperties.put("Click1", mouseEventHandler);
            specialBoxProperties.put("Click2", mouseEventHandler);
            specialBoxProperties.put("Click3", mouseEventHandler);
            specialBoxProperties.put("DoubleClick1", mouseEventHandler);
            specialBoxProperties.put("DoubleClick2", mouseEventHandler);
            specialBoxProperties.put("DoubleClick3", mouseEventHandler);

            specialBoxProperties.put("root", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.getRoot() == null) return null;
                        else if (b.parent == null) return b;
                        else return b.getRoot().getRootProxy();
                    } });

            specialBoxProperties.put("Minimized", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.parent == null && b.surface != null) return b.surface.minimized ? Boolean.TRUE : Boolean.FALSE;
                        else return null;
                    }
                    public void put(Box b, Object value) {
                        if (b.surface == null) return;
                        boolean val = stob(value);
                        if (b.parent == null && b.surface.minimized != val) b.surface.setMinimized(val);
                    }
                });

            specialBoxProperties.put("Maximized", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.parent == null && b.surface != null) return b.surface.maximized ? Boolean.TRUE : Boolean.FALSE;
                        else return null;
                    }
                    public void put(Box b, Object value) {
                        if (b.surface == null) return;
                        boolean val = stob(value);
                        if (b.parent == null && b.surface.maximized != val) b.surface.setMaximized(val);
                    }
                });

            specialBoxProperties.put("toback", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && stob(value) && b.surface != null) b.surface.toBack();
                    }
                });

            specialBoxProperties.put("tofront", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && stob(value) && b.surface != null) b.surface.toFront();
                    }
                });

            //#repeat hscar/vscar
            specialBoxProperties.put("hscar", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && b.surface != null) {
                            b.surface.hscar = stoi(value);
                            b.surface.dirty(0, 0, b.width, b.height);
                            b.surface.Refresh();
                        }
                    }
                });
            //#end

            specialBoxProperties.put("Close", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        if (b.parent == null && b.surface != null) b.surface.dispose(true);
                    }
                });

            // these are all do-nothings; just to prevent space from getting taken up in the params Hash.
            specialBoxProperties.put("KeyPressed", new SpecialBoxProperty());   // FIXME should cascade
            specialBoxProperties.put("KeyReleased", new SpecialBoxProperty());  // FIXME should cascade
            specialBoxProperties.put("PosChange", new SpecialBoxProperty());
            specialBoxProperties.put("SizeChange", new SpecialBoxProperty());

            //#repeat hpad/vpad 
            specialBoxProperties.put("hpad", new SpecialBoxProperty() {
                    public Object get(Box b) {
                        if (b.redirect == null) return new Integer(0);
                        if (b.redirect != b) return get(b.redirect);
                        return new Integer(b.hpad);
                    }
                    public void put(Box b, Object value) {
                        if (b.redirect == null) return;
                        if (b.redirect != b) { put(b.redirect, value); return; }
                        int newval = stoi(value);
                        if (newval == b.hpad) return;
                        b.hpad = newval;
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end

            //#repeat minwidth/minheight maxwidth/maxheight
            specialBoxProperties.put("minwidth", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.minwidth); }
                    public void put(Box b, Object value) {
                        if (stoi(value) == b.minwidth) return;
                        b.minwidth = stoi(value);
                        MARK_FOR_REFLOW_b;
                    }
                });
            specialBoxProperties.put("maxwidth", new SpecialBoxProperty() {
                    public Object get(Box b) { return new Integer(b.maxwidth); }
                    public void put(Box b, Object value) {
                        if (stoi(value) == b.maxwidth) return;
                        b.maxwidth = stoi(value);
                        MARK_FOR_REFLOW_b;
                    }
                });
            //#end

            specialBoxProperties.put("redirect", new SpecialBoxProperty() {
                    public void put(Box b, Object value) { }
                    public Object get(Box b) {
                        if (b.redirect == null) return null;
                        if (b.redirect == b) return Boolean.TRUE;
                        return get(b.redirect);
                    }
                });

            /*
            // FIXME: need to be able to read this back
            specialBoxProperties.put("titlebar", new SpecialBoxProperty() {
                    public void put(Box b, Object value) { surface.setTitleBarText(value.toString()); }
                    public Object get(Box b) { return b.ti; }
                });

            // FIXME: need to be able to read this back
            specialBoxProperties.put("icon", new SpecialBoxProperty() {
                    public void put(Box b, Object value) {
                        Picture pic = Box.getPicture(value.toString());
                        if (pic != null) surface.setIcon(pic);
                        else if (Log.on) Log.log(this, "unable to load icon " + value);
                    }
                    public Object get(Box b) { return b.id; }
                });
            */
        }

        
        /** helper that converts a String to a boolean according to JavaScript coercion rules */
        public static boolean stob(Object o) {
            if (o == null) return false;
            return Boolean.TRUE.equals(o) || "true".equals(o);
        }

        
    }
        /** helper that converts a String to an int according to JavaScript coercion rules */
        public static int stoi(Object o) {
            if (o == null) return 0;
            if (o instanceof Integer) return ((Integer)o).intValue();

            String s;
            if (!(o instanceof String)) s = o.toString();
            else s = (String)o;

            try { return Integer.parseInt(s.indexOf('.') == -1 ? s : s.substring(0, s.indexOf('.'))); }
            catch (NumberFormatException e) { return 0; }
        }
}
        

    
